/******************************************************************************

 This file is part of FDD3000 Emulator version 0.1.6

 Copyright (C) 2013,2014 by Sławomir Szczyrba <sszczyrba@gmail.com>

 FDD300Emu is free software: you can redistribute it and/or modify it under
 the terms of the GNU Lesser General Public License as published by the Free
 Software Foundation, either version 2 of the License, or (at your option) any
 later version.

 This program is distributed in the hope that it will be useful, but WITHOUT
 ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
 details.

 You should have received a copy of the GNU Lesser General Public License
 along with this program. If not, see <http://www.gnu.org/licenses/lgpl.html>

******************************************************************************/

#include <QDebug>
#include <QFile>
#include <QFileInfo>
#include <QFileDialog>
#include <QInputDialog>
#include <QMessageBox>
#include <QImage>

#include "diskImage.h"

diskImage::diskImage(QWidget *parent, QObject *gui, int id) :
    QLabel(parent)
{
    image=new QByteArray("");
    ID=id;
    logger=gui;
    loaded=0;
    dirty=0;
    side=0;
    system=0;
    rotate=rotateCnt=0;
    spinTimer   = new QTimer(this);
    rotateTimer = new QTimer(this);
    setGeometry(0,0,100,100);
    setContextMenuPolicy(Qt::CustomContextMenu);
    connect(this, SIGNAL(customContextMenuRequested(const QPoint&)), this, SLOT(contextMenu(const QPoint&)));
    connect(this, SIGNAL(message(QString)), logger, SLOT(message(QString)));
    connect(spinTimer,SIGNAL(timeout()),this,SLOT(stopSpin()));
    connect(rotateTimer,SIGNAL(timeout()),this,SLOT(rotateSpin()));

    connect(this,SIGNAL(spinStop()),this,SLOT(stopSpin()));
    connect(this,SIGNAL(spinStart()),this,SLOT(spin()));

    popup.addAction(tr("Save"));
    popup.actions().last()->setData(0);
    popup.addAction(tr("Save As"));
    popup.actions().last()->setData(1);
    popup.addSeparator();
    popup.addAction(tr("Insert disk"));
    popup.actions().last()->setData(2);
    popup.addAction(tr("Insert empty disk 160k"));
    popup.actions().last()->setData(3);
    popup.addAction(tr("Insert empty disk 640k"));
    popup.actions().last()->setData(4);
    popup.addSeparator();
    popup.addAction(tr("Rename"));
    popup.actions().last()->setData(5);
}

diskImage::~diskImage () {
    delete image;
}

int diskImage::insertEmpty(int fmt) {
    int size=163840;
    int now = QDateTime::currentMSecsSinceEpoch()/1000;
    if(loaded) { ejectImage(saveAuto); }
    switch (fmt) {
        case FDD160K: size=163840; format=FDD160K;trackJump=4096;lastTrack=40;break;
        case FDD320K: size=327680; format=FDD320K;trackJump=8192;lastTrack=40;break;
        case FDD640K: size=655360; format=FDD640K;trackJump=8192;lastTrack=80;break;
    default:    return dskUnknownFormat;
    }
    image->fill(0xe5, size);
    if(fmt!=FDD320K){
        char nul[32]={0};
        image->replace(0,16384,(const char*)diskTosA2,16384);
        image->replace(16384,32,nul,32);
        image->replace(16384,1,"\xff",1);
//        rename(tr("ANEWDISK"));
        if (fmt==FDD640K) { *(image->data()+0x0e)=1; }
    }
    loaded=1;
    protect=0;
    trackOffset=0;
    imageName="dsk"+QString::number(now,16)+".img";
    system=sysTOSv2;
    rename(tr("ANEWDISK"));
    emit(spinStart());
    return dskOK;
}

int diskImage::loadIMG(QString name){
    QFile dsk(name);

    if (!dsk.open(QIODevice::ReadOnly)) { return dskCannotOpen; }
    image->resize(0);
    image->append( dsk.readAll() );
    if(image->size()!=dsk.size()) {  return dskLoadError; }
    loaded=1;
    trackOffset=0;
    protect=0;
    dirty=0;
    imageName=name;
    switch (dsk.size()) {
        case 163840: format=FDD160K;trackJump=4096;lastTrack=40;break;
        case 327680: format=FDD320K;trackJump=8192;lastTrack=40;break;
        case 655360: format=FDD640K;trackJump=8192;lastTrack=80;break;
    }
    systemGuess();
    emit(spinStart());
    return dskOK;
}

int diskImage::loadDSK(QString name) {
    int tracks,sides,size,i;
    QString info="";
    QFile dsk(name);

    if (!dsk.open(QIODevice::ReadOnly)) return dskCannotOpen;
    image->resize(0);
    image->append(dsk.read(256));
    image->data()[0x15]=0;
    info.append(QString(image->data()));
    if (info=="MV - CPCEMU Disk-File") {
        tracks=image->at(0x30);
        sides=image->at(0x31);
        size=image->at(0x33);
        size=size*256+image->at(0x32);
    } else if (info=="EXTENDED CPC DSK File") {
        tracks=image->at(0x30);
        sides=image->at(0x31);
        size=image->at(0x34);
        size=size*256;
    } else { return dskDskNotSupported; }
    if(size != 4352) { return dskWrongTrackSize; }
    if((tracks==40)&&(sides==1)) {format=FDD160K;trackJump=4096;lastTrack=40; }
    if((tracks==40)&&(sides==2)) {format=FDD320K;trackJump=8192;lastTrack=40; }
    if((tracks==80)&&(sides==1)) {return dskWrongGeometry; }
    if((tracks==80)&&(sides==2)) {format=FDD640K;trackJump=8192;lastTrack=80; }
    image->resize(0);
    QByteArray tmp(dsk.size(),0x00);
    if((dsk.size()-256)!=dsk.read(tmp.data(),dsk.size()-256)) {return dskLoadError; }
    for(i=0; i<tracks; i++) {image->append(tmp.mid(0x100+i*4352),4096); }
    if (image->size()!=(lastTrack*trackJump)){ return dskImportError;}
    loaded=1;
    trackOffset=0;
    protect=0;
    dirty=0;
    imageName=name;
    systemGuess();
    emit(spinStart());
    return dskOK;
}

int diskImage::openImage(QString name){
    QFileInfo info(name);
    QString   ext=info.suffix();
    if(!info.exists()) {return dskFileNotfound;}
    if (ext.toUpper()=="IMG") { return loadIMG(name); }
    if (ext.toUpper()=="DSK") { return loadDSK(name); }
    /* unknown suffix. check lengths... */
    if ((info.size()==163840) || (info.size()==327680) || (info.size()==655360)){ return loadIMG(name); }
    if ((info.size()==174336) || (info.size()==348416) || (info.size()==696576)){ return loadDSK(name); }
    return dskUnknownFormat;
}

int diskImage::openImage(void){
    QString fileName = QFileDialog::getOpenFileName(
                NULL, tr("Select disk image"),
                ".",
                tr("image (*.img *.dsk)"),
                0,
                QFileDialog::DontUseNativeDialog);
    return openImage(fileName);
}

int diskImage::saveIMG(QString name) {
    QFile img(name);

    img.open(QIODevice::WriteOnly);
    img.resize(0);
    img.write(*image);
    img.flush();
    img.close();
    imageName=name;
    emit(spinStart());
    return dskOK;
}

int diskImage::saveDSK(QString name) {
    QFile      file(name);
    QByteArray header((const char*)dskHeader,256);
    int track, side, sec;

    file.open(QIODevice::WriteOnly);
    file.resize(0);
    if(format == FDD640K) {
        for (track=92;track<132;track++) { header.data()[track]=0x11; }
    }
    file.write(header);


    for (track=0; track<lastTrack; track++) {
        side=0;
        header.resize(0);
        header.append((const char*)trkHeader,256);
        header.data()[0x10]=(char)track&0xff;
        header.data()[0x11]=(char)side&0xff;
        for(sec=0;sec<16;sec++) {
            header.data()[0x18+sec*8]=(char)track&0xff;
            header.data()[0x19+sec*8]=(char)side&0xff;
        }
        file.write(header);
        file.write(image->mid(track*trackJump,4096));
        if(format != FDD160K) {
            side=1;
            header.resize(0);
            header.append((const char*)trkHeader,256);
            header.data()[0x10]=(char)track&0xff;
            header.data()[0x11]=(char)side&0xff;
            for(sec=0;sec<16;sec++) {
                header.data()[0x18+sec*8]=(char)track&0xff;
                header.data()[0x19+sec*8]=(char)side&0xff;
            }
            file.write(header);
            file.write(image->mid(4096+track*trackJump,4096));
        }
    }
    file.flush();
    file.close();
    emit(spinStart());
    imageName=name;
    return dskOK;
}

int diskImage::closeImage(QString name){
    QFileInfo info;
    if(name==""){ name = imageName;}
    info.setFile(name);
    if (info.suffix().toUpper()=="IMG") { return saveIMG(name); }
    if (info.suffix().toUpper()=="DSK") { return saveDSK(name); }
    return saveIMG(name+".img");
}

int diskImage::closeImage(int mode){
    QString fileName = "";
    switch(mode) {
    case saveAsk :fileName = QFileDialog::getSaveFileName(
                    NULL, tr("Save disk image"),
                    ".",
                    tr("image (*.img *.dsk)"),
                    0,
                    QFileDialog::DontUseNativeDialog);
                    /*,
                    0,
                    QFileDialog::DontConfirmOverwrite);*/
    case saveForce: closeImage(fileName); break;
    case saveAuto:  if(autosave && dirty) { closeImage(fileName); } break;
    }
    return dskDoNotSave;
}

int diskImage::ejectImage(int mode){
    int err;

    if(loaded) {
//        err=
        closeImage(mode);
//        qDebug()<<err;
//        if (err) { return err; } else { loaded=0; return dskOK; }
        loaded=0;
        tosName="";
        system=sysNone;
        return dskOK;
    } else {
        err=openImage();
        systemGuess();
        return err;
    }
    return dskOK;
}

int diskImage::swapSide(void){
    if(format==FDD320K) {
        if(side) { trackOffset=0; side=0;}
        else {trackOffset=0/*4096*/; side=1; }
        emit(spinStart());
    }
    return dskOK;
}
/*********************** sector / track ******************/

int diskImage::loadSector(int side, int track, int sector, QByteArray * buf){
//    qDebug()<<"id "<<ID<<" side "<<side<< " track "<<track<< " sector "<<sector;
    if(!loaded) { return dskNoDisk; }
    if(side && (format==FDD160K)){ return dskSingleSide; }
    if(track>= lastTrack) { return dskTrackTooHigh; }
    if(sector>15) { return dskSectorTooHigh; }
    if(track<0){  track=0; }
    if(sector<0){ sector=0; }
    buf->resize(0);
    buf->append(image->mid(trackOffset+side*4096+track*trackJump+sector*256,256));
    emit(spinStart());
    return dskOK;
}

int diskImage::loadTrack(int side, int track, QByteArray * buf){
    if(!loaded) { return dskNoDisk; }
    if(side && (format==FDD160K)){ return dskSingleSide; }
    if(track>= lastTrack) { return dskTrackTooHigh; }
    if(track<0){ track=0; }
    buf->resize(0);
    buf->append(image->mid(trackOffset+side*4096+track*trackJump,4096));
    emit(spinStart());
    return dskOK;
}

int diskImage::saveSector(int side, int track, int sector, QByteArray * buf){
//    qDebug()<<"id "<<ID<<" side "<<side<< " track "<<track<< " sector "<<sector;
    if(!loaded) { return dskNoDisk; }
    if(protect){ return dskReadOnly; }
    if(side && (format==FDD160K)){ return dskSingleSide; }
    if(track>= lastTrack) { return dskTrackTooHigh; }
    if(track<0){ track=0; }
    if(sector>15) { return dskSectorTooHigh; }
    if(sector<0){ sector=0; }
    if(buf->size()!=256){ buf->resize(256); } /* tak na wszelki wypadek*/
    image->replace(trackOffset+side*4096+track*trackJump+sector*256,256,*buf);
    emit(spinStart());
    return dskOK;
}

int diskImage::saveTrack(int side, int track, QByteArray * buf){
    if(!loaded) { return dskNoDisk; }
    if(protect){ return dskReadOnly; }
    if(side && (format==FDD160K)){ return dskSingleSide; }
    if(track>= lastTrack) { return dskTrackTooHigh; }
    if(track<0){ track=0; }
    if(buf->size()!=4096){ buf->resize(4096); } /* tak na wszelki wypadek*/
    image->replace(trackOffset+side*4096+track*trackJump,4096,*buf);
    emit(spinStart());
    return dskOK;
}

void diskImage::systemGuess() {
    QByteArray track,data;
    QString name;
    int i;

    system=sysUnknown;
    tosName="UNKNOWN";
    if(!loaded) {
        qDebug()<<"Not loaded...";
        return;
    }
    data.clear();
    for (i=0; i<6; i++){/* qDebug()<<"Load track "<<i; qDebug()<<*/loadTrack(side,i,&track); data.append(track); }
/*
    QFile fram("image.dat");
    fram.open(QIODevice::WriteOnly|QIODevice::Truncate);
    fram.write(data);
    fram.close();
*/
    if (data.mid(0x012c,16)=="DIGITAL RESEARCH") { /* qDebug() << "CPM"; */ system=sysCPM; tosName="CP/M DISK"; return; }
    if (data.mid(0x0000, 8)=="\xE5\xE5\xE5\xE5\xE5\xE5\xE5\xE5"){ /* qDebug() << "EMPTY"; */ system=sysNone; tosName="EMPTYDSK"; return;}
    if (data.mid(0x3f9f,13)=="MaxCurRemName")    {
//        qDebug() << "TOS";
        system=sysTOSv2;
        name=data.mid(0x4001,8);
        if(name[0]!=0xe5){ tosName=name.trimmed(); }
//        qDebug() << tosName;
    }
//    qDebug() << "systemGuess end";
}

/*
QString diskImage::tosName(void) {
    QString name="";
    //const QByteArray im=image;
    if(loaded) {
//        qDebug() << "loaded";
//        QFile out("buffer.dat");
//        out.open(QIODevice::WriteOnly|QIODevice::Truncate);
//        out.write(*image);
//        out.close();
        if (system == sysCPM)  { return "CP/M DISK"; }
        if (system == sysTOSv2){
            name=image->mid(0x4001,8);
            if(name[0]!=0xe5){ return name.trimmed(); }
        }
        return "UNKNOWN";
*/
/*
        if(image->data()[0x4000]==(char)0xff){
            name=image->mid(0x4001,8);
        } else if((image->data()[0x2000]==(char)0) ||(image->data()[0x2000]==(char)0xff) ){
            name = image->mid(0x2001,8);
        }
//        qDebug()<<"Fname:"+name;
        if(name[0]!=0xe5){ return name.trimmed(); }
*/
/*
        if(image->at(0x4000)==(QChar)0xff){
            name=image->mid(0x4001,8);
        } else if((image->at(0x2000)==(QChar)0) ||(image->at(0x2000)==(QChar)0xff) ){
            name = image->mid(0x2001,8);
        }
//        qDebug()<<"Fname:"+name;
        if(name[0]!=0xe5){ return name.trimmed(); }
*/

/*
    }
    return "";
}
*/

void diskImage::rename(QString name) {
    int i;
    if (system==sysTOSv2) {
        QString allowed=" ABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789_-";
        // if(cpmImage) { return; } // TODO :)
        name=name.toUpper().simplified().trimmed();
        name=name.replace(' ','_');
        for(i=name.length()-1; i>=0; i--) {
            if(!allowed.indexOf(name.at(i)) ) { name.remove(i,1); }
        }
        name+="         ";
        if(!name.trimmed().length()){ return; }
        tosName=name.left(8);
        name=tosName+"DIR";
        image->replace(0x4001,11,name.toLatin1(),11);

//        qDebug()<<"NewName: "<<name;
    }
}

int diskImage::isLoaded(void) { return loaded; }

int diskImage::tracks(void) { return lastTrack; }

int diskImage::isSwapable(void) { return format == FDD320K; }

int diskImage::isProtected(void) { return protect; }

//void diskImage::setID(int val) { ID=val; }

void diskImage::setProtection(int ro) { protect=ro; }

void diskImage::setAutosave(int a) { autosave = a; }

/******************* label stuff **************************/


const QString diskImage::icoNames[]={
    ":gfx/floppy/no_floppy.png",  //0
    ":gfx/floppy/floppy.png",     //1
    ":gfx/floppy/floppy_a.png",   //2
    ":gfx/floppy/floppy_b.png",   //3
    ":gfx/floppy/rotor_0.png",    //4
    ":gfx/floppy/rotor_1.png",    //5
    ":gfx/floppy/rotor_2.png",    //6
    ":gfx/floppy/rotor_3.png",    //7
};

void diskImage::spin() {
    if(!rotate) {
        rotateTimer->start(100);
        rotate=1;
    }
    spinTimer->start(1000);
}

void diskImage::paintEvent(QPaintEvent *event) {
    QPainter paint;
    QString name;
    int ico,i;

    if (icons[0].isNull()) {
        for(i=0;i<8;i++){ icons[i]=QPixmap(icoNames[i]); }
    }

    if (!loaded){ ico=icoEmpty; } else {
        if (format==FDD320K) { ico = icoSideA + side; }
        else { ico = icoLoaded; }
    }

//    try{
        setPixmap(icons[ico]);
        paint.begin(this);
        paint.drawText(10,25,tosName);
        if(rotate) { paint.drawPixmap(39,43,22,22,icons[icoRotate+rotateCnt]); }
        paint.end();
//    }catch(...){
//        qDebug()<<"Jebudup @ paint event";
//    }

    QLabel::paintEvent(event);
//    qDebug()<<"paint";
}

void diskImage::mousePressEvent(QMouseEvent *event) {
  if (event->button() == Qt::LeftButton) {

    if( isSwapable() && QRect(8,72,20,20).contains(event->pos())) {
        swapSide();
        systemGuess();
        message(tr("Swapped."));
        return;
    }
    ejectImage(saveAuto);
    update();
  }
//  QWidget::mousePressEvent(event);
}

void diskImage::stopSpin(void) {
//    if(loaded) { setIcon(icoLoaded); } else { setIcon(icoEmpty); }
    rotate=0;
    spinTimer->stop();
    rotateTimer->stop();
}

void diskImage::rotateSpin(void) {
    rotateCnt=(rotateCnt+1)%4;
    update();
}

void diskImage::contextMenu(const QPoint& pos) {
    QString name;
    QInputDialog dlg;
    if(loaded) {
        popup.actions().at(0)->setEnabled(true);
        popup.actions().at(1)->setEnabled(true);
        popup.actions().at(7)->setEnabled(true);
    } else {
        popup.actions().at(0)->setEnabled(false);
        popup.actions().at(1)->setEnabled(false);
        popup.actions().at(7)->setEnabled(false);
    }
    if (system!=sysTOSv2) { popup.actions().at(7)->setEnabled(false); }
    QAction * selected = popup.exec(this->mapToGlobal(pos));
    if (selected) {
        switch(selected->data().toInt()){
        case 0 : closeImage(saveForce); break;
        case 1 : closeImage(saveAsk); break;
        case 2 : openImage(); break;
        case 3 : insertEmpty(FDD160K); /*rename(tr("ANEWDISK"));*/ break;
        case 4 : insertEmpty(FDD640K); /*rename(tr("ANEWDISK"));*/ break;
        case 5 : dlg.setStyleSheet("QInputDialog{ background:solid;}");
            name=QInputDialog::getText(this,tr("Rename disk"),tr("New name:"));
            if(name.length()) { rename(name); };
            break;
        }
    }
}
